SystemOrganization addCategory: #'Zinc-FileSystem'!

Object subclass: #ZnFileSystemUtils
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-FileSystem'!

!ZnFileSystemUtils commentStamp: 'SvenVanCaekenberghe 7/11/2018 16:29' prior: 0!
I am ZnFileSystemUtils.
I help is abstracting over API differences pre and post the introduction of FileSystem in Pharo 2.0 as well as pre and post the important stream and file changes midway in Pharo 7.0

Part of Zinc HTTP Components.!

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
baseNameFor: fileName
	^ fileName asFileReference basename ! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 9/27/2018 14:42'!
binaryReadStreamFor: fileName
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryReadStream)
		ifTrue: [ fileReference binaryReadStream ]
		ifFalse: [ fileReference readStream binary ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:54'!
binaryWriteStreamFor: fileName
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryWriteStream)
		ifTrue: [ fileReference binaryWriteStream ]
		ifFalse: [ fileReference writeStream binary ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:54'!
binaryWriteStreamFor: fileName do: block
	| fileReference |
	fileReference := fileName asFileReference.
	^ (fileReference respondsTo: #binaryWriteStreamDo:ifPresent:)
		ifTrue: [
			fileName asFileReference 
				binaryWriteStreamDo: block 
				ifPresent: [ FileExists signalWith: fileReference ] ]
		ifFalse: [ 
			fileReference isFile 
				ifTrue: [ FileExists signalWith: fileReference ] 
				ifFalse: [ | binaryStream |
					binaryStream := self binaryWriteStreamFor: fileName.
					[ block value: binaryStream ] ensure: [ binaryStream close ] ] ]! !

!ZnFileSystemUtils class methodsFor: 'defaults' stamp: ''!
defaultDirectory
	^ FileSystem disk workingDirectory! !

!ZnFileSystemUtils class methodsFor: 'defaults' stamp: ''!
defaultDirectoryPath
	^ self defaultDirectory fullName! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
delete: path
	^ path asFileReference delete! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
deleteIfExists: path
	^ (self exists: path)
		ifTrue: [ self delete: path. true ]
		ifFalse: [ false ]! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
directory: path
	^ path asFileReference! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
downloadPathTo: directoryOrFileName for: baseName
	| downloadPath |
	downloadPath := directoryOrFileName asFileReference.
	^ downloadPath isDirectory
		ifTrue: [
			baseName ifNil: [ 
				self error: 'Not enough information to construct a download path' ]. 
			(downloadPath / baseName) fullName ]
		ifFalse: [ 
			directoryOrFileName ]
	! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
exists: path
	^ path asFileReference exists! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
extensionFor: fileName
	^ fileName asFileReference extension! !

!ZnFileSystemUtils class methodsFor: 'queries' stamp: ''!
fileSizeFor: fileName
	^ fileName asFileReference entry size! !

!ZnFileSystemUtils class methodsFor: 'paths' stamp: ''!
fullNameFor: fileName
	^ fileName asFileReference fullName! !

!ZnFileSystemUtils class methodsFor: 'queries' stamp: ''!
modificationTimeFor: path
	^ path asFileReference entry modificationTime ! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 9/27/2018 14:43'!
readStreamFor: fileName
	^ fileName asFileReference readStream! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:49'!
readStreamFor: fileName do: block
	^ fileName asFileReference 
		readStreamDo: block
		ifAbsent: [ FileDoesNotExist signalWith: fileName asFileReference ]! !

!ZnFileSystemUtils class methodsFor: 'streams' stamp: 'SvenVanCaekenberghe 11/1/2018 23:51'!
writeStreamFor: fileName do: block
	^ fileName asFileReference 
		writeStreamDo: block 
		ifPresent: [ FileExists signalWith: fileName asFileReference ]! !

Object subclass: #ZnMonticelloServerDelegate
	instanceVariableNames: 'directory'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-FileSystem'!

!ZnMonticelloServerDelegate commentStamp: '' prior: 0!
I am ZnMonticelloServerDelegate.
I am a proof of concept implementation of a basic Monticello server.
I function as a delegate for ZnServer.

ZnServer startDefaultOn: 1701.
ZnServer default delegate: ((ZnMonticelloServerDelegate new) 
									directory: '/tmp/monticello'; 
									yourself).

Part of Zinc HTTP Components.!

!ZnMonticelloServerDelegate methodsFor: 'accessing' stamp: ''!
directory
	^ directory ! !

!ZnMonticelloServerDelegate methodsFor: 'accessing' stamp: ''!
directory: object
	directory := object asFileReference! !

!ZnMonticelloServerDelegate methodsFor: 'private' stamp: ''!
handleGetMczEntry: request
	| name file stream entity |
	name := request uri firstPathSegment.
	file := self directory / name.
	^ ((self isValidMczName: name) and: [ file exists ]) 
		ifTrue: [ |  |
			(stream := file readStream)
				binary.
			(entity := ZnStreamingEntity type: ZnMimeType applicationOctetStream length: file size)
				stream: stream.
			ZnResponse ok: entity ]
		ifFalse: [ ZnResponse notFound: request uri ]! !

!ZnMonticelloServerDelegate methodsFor: 'private' stamp: ''!
handleListMczEntries: request
	^ ZnResponse ok: (ZnEntity html: self repositoryListing)! !

!ZnMonticelloServerDelegate methodsFor: 'private' stamp: ''!
handlePutMczEntry: request
	| name file |
	name := request uri firstPathSegment.
	file := self directory / name.
	^ ((self isValidMczName: name) and: [ file exists not ]) 
		ifTrue: [
			file writeStreamDo: [ :stream |
				stream binary.
				request entity writeOn: stream ].
			ZnResponse created: request uri ]
		ifFalse: [ ZnResponse badRequest: request ]! !

!ZnMonticelloServerDelegate methodsFor: 'public' stamp: ''!
handleRequest: request
	(request method = #GET)
		ifTrue: [
			^ request uri isSlash
				ifTrue: [ self handleListMczEntries: request ]
				ifFalse: [ self handleGetMczEntry: request ] ].
	^ (request method = #PUT) 
		ifTrue: [ self handlePutMczEntry: request ]
		ifFalse: [ ZnResponse badRequest: request ]! !

!ZnMonticelloServerDelegate methodsFor: 'private' stamp: ''!
isValidMczName: string
	^ string isString 
		and: [ string isEmpty not
			and: [ (string endsWith: '.mcz')
				and: [ (string includes: $/) not ] ] ]! !

!ZnMonticelloServerDelegate methodsFor: 'accessing' stamp: ''!
mczEntries
	^ (self directory fileNames select: [ :each | each endsWith: '.mcz' ])
		asSortedCollection! !

!ZnMonticelloServerDelegate methodsFor: 'private' stamp: ''!
repositoryListing
	^ ZnHtmlOutputStream streamContents: [ :html |
		html page: 'Monticello Repository' do: [
			html tag: #ul do: [ 
				self mczEntries do: [ :each |
					html tag: #li do: [ 
						html 
							tag: #a 
							attributes: { #href. each } 
							with: each ] ] ] ] ]! !

!ZnMonticelloServerDelegate methodsFor: 'public' stamp: ''!
value: request
	"I implement the generic #value: message as equivalent to #handleRequest:"
	
	^ self handleRequest: request! !

Object subclass: #ZnStaticFileServerDelegate
	instanceVariableNames: 'prefix directory mimeTypeExpirations'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Zinc-FileSystem'!

!ZnStaticFileServerDelegate commentStamp: '' prior: 0!
I am ZnStaticFileServerDelegate.
I am a simple proof of concept implementation of a web server serving static files.
I handle urls with an optional prefix as requests for files in a directory.
I serve index.html or index.htm when a directory is requested and these files exist.
I do a redirect when a path that is not does not end with a / refers to directory.
I function as a delegate for ZnServer.

ZnServer startDefaultOn: 1701.
ZnServer default delegate: ((ZnStaticFileServerDelegate new) 
									prefixFromString: 'apple/macosx'; 
									directory: '/Library/WebServer/Documents' asFileReference; 
									yourself).

Part of Zinc HTTP Components.!

!ZnStaticFileServerDelegate class methodsFor: 'accessing' stamp: ''!
defaultMimeTypeExpirations
	"Return a default dictionary mapping ZnMimeTypes to their expiration time in seconds.
	Missing entries will have no expiration time."

	| dict |
	dict := Dictionary new.
	dict at: ZnMimeType textCss put: 30 days asSeconds.
	dict at: ZnMimeType textJavascript put: 30 days asSeconds.
	dict at: ZnMimeType textPlain put: 30 days asSeconds.
	dict at: ZnMimeType textHtml put: 30 days asSeconds.
	dict at: ZnMimeType imageGif put: 30 days asSeconds.
	dict at: ZnMimeType imageJpeg put: 30 days asSeconds.
	dict at: ZnMimeType imagePng put: 30 days asSeconds.
	dict at: ZnMimeType applicationJavascript put: 30 days asSeconds.
	^ dict! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
actualFilenameFor: uri
	| subElements subDir entry |
	(uri isEmpty and: [ self prefix isEmpty ]) ifTrue: [ ^ self indexFileIn: self directory ].
	(self prefix isEmpty or: [ uri isEmpty not and: [ uri pathSegments beginsWith: self prefix ] ]) ifFalse: [ ^ nil ].
	subElements := (uri pathSegments allButFirst: self prefix size) reject: [ :each | each = #/ ].
	subDir := subElements allButLast inject: self directory into: [ :parent :sub | | file |
		(file := parent / sub) exists
			ifTrue: [ file ]
			ifFalse: [ ^ nil ] ]. 
	subElements isEmpty
		ifTrue: [ entry := subDir entry ] 
		ifFalse: [ | file |
			 (file := subDir / subElements last) exists 
				ifTrue: [ entry := file entry ] 
				ifFalse: [ ^ nil ] ].
	^ entry isDirectory 
		ifTrue: [ self indexFileIn: entry reference ]
		ifFalse: [ entry reference fullName ]! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
directory
	"The directory whose files I am serving"
	
	^ directory ! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
directory: fileDirectory
	"Set the directory whose files I should be serving"

	directory := fileDirectory! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
directoryRedirectFor: uri
	| directoryUri |
	directoryUri := uri copy addPathSegment: #/.  
	^ ZnResponse redirect: directoryUri ! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
expirationDateFor: aMimeType
	| expiration |
	expiration := self mimeTypeExpirations at: aMimeType ifAbsent: [ ^ nil ].
	^ ZnUtils httpDate: (DateAndTime now asUTC + expiration seconds)! !

!ZnStaticFileServerDelegate methodsFor: 'public' stamp: ''!
handleRequest: request
	"Server delegate entry point"

	| actualFilename |
	(#(#GET #HEAD) includes: request method)
		ifFalse: [ ^ ZnResponse methodNotAllowed: request ].
	actualFilename := self actualFilenameFor: request uri.
	^ actualFilename notNil
		ifTrue: [ 
			(self redirectNeededFor: request uri actualFilename: actualFilename)
				ifTrue: [ self directoryRedirectFor: request uri ]
				ifFalse: [ self responseForFile: actualFilename fromRequest: request ] ]
		ifFalse: [ ZnResponse notFound: request uri ]! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
indexFileIn: fileDirectory
	#( 'index.html' 'index.htm' ) do: [ :each | | file |
			(file := fileDirectory / each) exists
				ifTrue: [ 
					^ file fullName ] ].
	^ nil
! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
maxAgeFor: aMimeType
	| expiration |
	expiration := self mimeTypeExpirations at: aMimeType ifAbsent: [ ^ nil ].
	^ 'max-age=', expiration asString! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
mimeTypeExpirations
	"Return a dictionary mapping ZnMimeTypes to their expiration time in seconds.
	Missing entries will have no expiration time."

	^ mimeTypeExpirations ifNil: [ mimeTypeExpirations := Dictionary new ]! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
mimeTypeExpirations: aDictionary
	"Set the mapping of ZnMimeTypes to their expiration time in seconds to aDictionary.
	Missing entries will have no expiration time."

	mimeTypeExpirations := aDictionary! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
prefix
	"The path prefix collection under which I am serving files"
	
	^ prefix ifNil: [ prefix := #() ]! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
prefix: orderedCollection
	"Set the path prefix under which I should be serving files to orderedCollection containing path elements"
	
	prefix := orderedCollection! !

!ZnStaticFileServerDelegate methodsFor: 'accessing' stamp: ''!
prefixFromString: string
	"Set the path prefix under which I should be serving files to string,
	interpreting each /-separated token as a path element"

	self prefix: (string findTokens: '/')! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
redirectNeededFor: uri actualFilename: actualFilename
	uri isDirectoryPath ifTrue: [ ^ false ].  
	^ (actualFilename endsWith: uri lastPathSegment) not! !

!ZnStaticFileServerDelegate methodsFor: 'private' stamp: ''!
responseForFile: filename fromRequest: aRequest
	| file entry size time mimeType fileStream entity response modified |
	file := filename asFileReference.
	entry := file entry.
	size := entry size.
	time := entry modificationTime.
	modified := true.
	aRequest headers
		at: 'If-Modified-Since'
		ifPresent: [ modified := time > (ZnUtils parseHttpDate: (aRequest headers at: 'If-Modified-Since')) ].
	modified
		ifTrue: [ 
			mimeType := ZnMimeType forFilenameExtension: file extension.
			fileStream := file binaryReadStream.
			(entity := ZnStreamingEntity type: mimeType length: size) 
				stream: fileStream.
			(response := ZnResponse ok: entity) 
				headers at: 'Modification-Date' put: (ZnUtils httpDate: time).
			(self maxAgeFor: mimeType) 
				ifNotNil: [ :maxAge | response headers at: 'Cache-Control' put: maxAge ].
			(self expirationDateFor: mimeType)
				ifNotNil: [ :expirationDate | response headers at: 'Expires' put: expirationDate ] ]
		ifFalse: [ response := ZnResponse notModified ].
	^ response! !

!ZnStaticFileServerDelegate methodsFor: 'public' stamp: ''!
value: request
	"I implement the generic #value: message as equivalent to #handleRequest:"
	
	^ self handleRequest: request! !
